def test_server_runner(redis_server):
    """ Boot up a server, push some keys into it
    """
    assert redis_server.check_server_up()
    redis_server.api.set('foo', 'bar')
    assert redis_server.api.get('foo').decode('utf8') == 'bar'



"""
Docker server class implementation.
"""
from __future__ import absolute_import

import logging
import docker

from retry import retry
from pytest_server_fixtures import CONFIG
from .common import (ServerClass,
                     merge_dicts,
                     ServerFixtureNotRunningException,
                     ServerFixtureNotTerminatedException)

log = logging.getLogger(__name__)


class DockerServer(ServerClass):
    """Docker server class."""

    def __init__(self,
                 server_type,
                 cmd,
                 get_args,
                 env,
                 image,
                 labels={}):
        super(DockerServer, self).__init__(cmd, get_args, env)

        self._image = image
        self._labels = merge_dicts(labels, {
            'server-fixtures': 'docker-server-fixtures',
            'server-fixtures/server-type': server_type,
            'server-fixtures/session-id': CONFIG.session_id,
        })

        self._client = docker.from_env()
        self._container = None

    def launch(self):
        try:
            log.debug('Launching container')
            self._container = self._client.containers.run(
                image=self._image,
                name=self.name,
                command=[self._cmd] + self._get_args(),
                environment=self._env,
                labels=self._labels,
                detach=True,
                auto_remove=True,
            )
            self._wait_until_running()
            log.debug('Container is running at %s', self.hostname)
        except docker.errors.ImageNotFound as err:
            log.warning("Failed to start container, image %s not found", self._image)
            log.debug(err)
            raise
        except docker.errors.APIError as e:
            log.warning("Failed to start container: %s", e)
            raise

        self.start()

    def run(self):
        try:
            self._container.wait()
        except docker.errors.APIError as e:
            log.warning("Error while waiting for container: %s", e)
            log.debug(self._container.logs())

    def teardown(self):
        if not self._container:
            return

        try:
            # stopping container will also remove it as 'auto_remove' is set
            self._container.stop()
            self._wait_until_terminated()
        except docker.errors.APIError as e:
            log.warning("Error when stopping the container: %s", e)

    @property
    def is_running(self):
        if not self._container:
            return False

        return self._get_status() == 'running'

    @property
    def hostname(self):
        if not self.is_running:
            raise ServerFixtureNotRunningException()
        return self._container.attrs['NetworkSettings']['IPAddress']

    def _get_status(self):
        try:
            self._container.reload()
            return self._container.status
        except docker.errors.APIError as e:
            log.warning("Failed to get container status: %s", e)
            raise

    @retry(ServerFixtureNotRunningException,
           tries=28,
           delay=1,
           backoff=2,
           max_delay=10)
    def _wait_until_running(self):
        if not self.is_running:
            raise ServerFixtureNotRunningException()

    @retry(ServerFixtureNotTerminatedException,
           tries=28,
           delay=1,
           backoff=2,
           max_delay=10)
    def _wait_until_terminated(self):
        try:
            self._get_status()
        except docker.errors.APIError as e:
            if e.response.status_code == 404:
                return
            raise



from mock import create_autospec, sentinel, call, patch, Mock

from pytest_server_fixtures.base import TestServer as _TestServer  # So that pytest doesnt think this is a test case


def test_init():
    ws = Mock()
    with patch('pytest_shutil.workspace.Workspace.__init__', autospec=True) as init:
        ts = _TestServer(workspace=ws, delete=sentinel.delete,
                         port=sentinel.port, hostname=sentinel.hostname)
    assert init.call_args_list == [call(ts, workspace=ws, delete=sentinel.delete)]
    assert ts.hostname == sentinel.hostname
    assert ts.port == sentinel.port
    assert ts.dead is False

    # Silence teardown warnings
    ts.dead = True
    ts.workspace = ws


def test_kill():
    self = create_autospec(_TestServer, dead=False,
                           hostname=sentinel.hostname,
                           port=sentinel.port)
    self.run.side_effect = ['100\n', '', '']
    with patch('os.kill') as kill:
        with patch('socket.gethostbyname', return_value=sentinel.ip):
            _TestServer._find_and_kill(self, 2, sentinel.signal)
    assert self.run.call_args_list == [call("netstat -anp 2>/dev/null | grep sentinel.ip:sentinel.port "
                                            "| grep LISTEN | awk '{ print $7 }' | cut -d'/' -f1", capture=True, cd='/'),
                                       call("netstat -anp 2>/dev/null | grep sentinel.ip:sentinel.port "
                                            "| grep LISTEN | awk '{ print $7 }' | cut -d'/' -f1", capture=True, cd='/')]
    assert kill.call_args_list == [call(100, sentinel.signal)]



# Common setup.py code shared between all the projects in this repository
import sys
import os
import logging

from setuptools.command.test import test as TestCommand
from setuptools.command.egg_info import egg_info as EggInfoCommand


class PyTest(TestCommand):
    pytest_args = []
    src_dir = None

    def initialize_options(self):
        TestCommand.initialize_options(self)

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        global pytest_args
        logging.basicConfig(format='%(asctime)s %(levelname)s %(name)s %(message)s', level='DEBUG')

        # import here, cause outside the eggs aren't loaded
        import pytest

        self.pytest_args.extend(['--junitxml', 'junit.xml'])
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


class EggInfo(EggInfoCommand):
    """ Customisation of the package metadata creation. Changes are:
         - Save the test requirements into an extra called 'tests'
    """
    def run(self):
        if self.distribution.extras_require is None:
            self.distribution.extras_require = {}
        self.distribution.extras_require['tests'] = self.distribution.tests_require
        EggInfoCommand.run(self)


def common_setup(src_dir):
    this_dir = os.path.dirname(__file__)
    readme_file = os.path.join(this_dir, 'README.md')
    changelog_file = os.path.join(this_dir, 'CHANGES.md')
    version_file = os.path.join(this_dir, 'VERSION')

    # Convert Markdown to RST for PyPI
    try:
        import pypandoc
        long_description = pypandoc.convert(readme_file, 'rst')
        changelog = pypandoc.convert(changelog_file, 'rst')
    except (IOError, ImportError, OSError):
        long_description = open(readme_file).read()
        changelog = open(changelog_file).read()

    # Gather trailing arguments for pytest, this can't be done using setuptools' api
    if 'test' in sys.argv:
        PyTest.pytest_args = sys.argv[sys.argv.index('test') + 1:]
        if PyTest.pytest_args:
            sys.argv = sys.argv[:-len(PyTest.pytest_args)]
    PyTest.src_dir = src_dir

    return dict(
            # Version is shared between all the projects in this repo
            version=open(version_file).read().strip(),
            long_description='\n'.join((long_description, changelog)),
            url='https://github.com/manahl/pytest-plugins',
            license='MIT license',
            platforms=['unix', 'linux'],
            cmdclass={'test': PyTest, 'egg_info': EggInfo},
            setup_requires=['setuptools-git'],
            include_package_data=True
            )



'''
Created on 25 Apr 2012

@author: eeaston
'''
from __future__ import absolute_import

import os.path
import shutil

import pytest
import six

from pytest_server_fixtures import CONFIG
from pytest_fixture_config import yield_requires_config

from .http import HTTPTestServer


@pytest.yield_fixture(scope='session')
@yield_requires_config(CONFIG, ['jenkins_war', 'java_executable'])
def jenkins_server():
    """ Session-scoped Jenkins server instance

        Attributes
        ----------
        api (`jenkins.Jenkins`)  : python-jenkins client API connected to this server
        .. also inherits all attributes from the `workspace` fixture
    """
    with JenkinsTestServer() as p:
        p.start()
        yield p


@pytest.yield_fixture(scope='module')
@yield_requires_config(CONFIG, ['jenkins_war', 'java_executable'])
def jenkins_server_module():
    """ Module-scoped Jenkins server instance

        Attributes
        ----------
        api (`jenkins.Jenkins`)  : python-jenkins client API connected to this server
        .. also inherits all attributes from the `workspace` fixture
    """
    with JenkinsTestServer() as p:
        p.start()
        yield p


class JenkinsTestServer(HTTPTestServer):
    port_seed = 65533
    kill_retry_delay = 2

    def __init__(self, **kwargs):
        global jenkins
        try:
            import jenkins
        except ImportError:
            pytest.skip('python-jenkins not installed, skipping test')
        super(JenkinsTestServer, self).__init__(**kwargs)
        self.env = dict(JENKINS_HOME=self.workspace,
                        JENKINS_RUN=self.workspace / 'run',
                        # Use at most 1GB of RAM for the server
                        JAVA_ARGS='-Xms1G -Xmx1G',
                        RUN_STANDALONE='true',
                        JENKINS_LOG=self.workspace / 'jenkins.log',
                        )
        self.api = jenkins.Jenkins(self.uri)

    @property
    def run_cmd(self):
        if not CONFIG.jenkins_war:
            raise ValueError("jenkins_war missing from org config")

        return [CONFIG.java_executable,
                '-jar', CONFIG.jenkins_war,
                '--httpPort=%s' % self.port,
                '--httpListenAddress=%s' % self.hostname,
                '--ajp13Port=-1',
                '--webroot={0}'.format(self.workspace / 'run' / 'war'),
                ]

    def load_plugins(self, plugins_repo, plugins=None):
        """plugins_repo is the place from which the plugins can be copied to this jenskins instance
           is plugins is None, all plugins will be copied, else is should be a list of the plugin names
        """

        if not os.path.isdir(plugins_repo):
            raise ValueError('Plugin repository "%s" does not exist' % plugins_repo)

        # copy the plugins to the jenkins plugin directory
        available_plugins = dict(((os.path.splitext(os.path.basename(x))[0], os.path.join(plugins_repo, x))
                                  for x in os.listdir(plugins_repo) if x.endswith('.hpi')))

        if plugins is None:
            plugins = available_plugins.keys()
        else:
            if isinstance(plugins, six.string_types):
                plugins = [plugins]

            errors = []
            for p in plugins:
                if p not in available_plugins:
                    if p not in errors:
                        errors.append(p)
            if errors:
                if len(errors) == 1:
                    e = 'Plugin "%s" is not present in the repository' % errors[0]
                else:
                    e = 'Plugins %s are not present in the repository' % sorted(errors)
                raise ValueError(e)

        for p in plugins:
            tgt = os.path.join(self.plugins_dir, '%s.hpi' % p)
            shutil.copy(available_plugins[p], tgt)

    @property
    def plugins_dir(self):
        return os.path.normpath(os.path.join(self.workspace, 'plugins'))



"""
Implementation of how a server fixture will run.
"""
# flake8: noqa
from __future__ import absolute_import

def create_server(server_class, **kwargs):
    if server_class == 'thread':
        from .thread import ThreadServer
        return ThreadServer(
            cmd=kwargs["cmd_local"],
            get_args=kwargs["get_args"],
            env=kwargs["env"],
            workspace=kwargs["workspace"],
            cwd=kwargs["cwd"],
            random_hostname=kwargs["random_hostname"],
        )

    if server_class == 'docker':
        from .docker import DockerServer
        return DockerServer(
            server_type=kwargs["server_type"],
            cmd=kwargs["cmd"],
            get_args=kwargs["get_args"],
            env=kwargs["env"],
            image=kwargs["image"],
            labels=kwargs["labels"],
        )

    if server_class == 'kubernetes':
        from .kubernetes import KubernetesServer
        return KubernetesServer(
            server_type=kwargs["server_type"],
            cmd=kwargs["cmd"],
            get_args=kwargs["get_args"],
            env=kwargs["env"],
            image=kwargs["image"],
            labels=kwargs["labels"],
        )



from mock import sentinel, patch, Mock

from pytest_server_fixtures import CONFIG
from pytest_server_fixtures.serverclass.thread import ThreadServer


@patch('pytest_server_fixtures.serverclass.thread.ServerClass.__init__')
def test_init(mock_init):
    ts = ThreadServer(sentinel.cmd,
                      sentinel.get_args,
                      sentinel.env,
                      sentinel.workspace,
                      cwd=sentinel.cwd,
                      random_hostname=False)

    mock_init.assert_called_with(sentinel.cmd,
                                 sentinel.get_args,
                                 sentinel.env)
    assert ts._workspace == sentinel.workspace
    assert ts._cwd == sentinel.cwd


@patch('pytest_server_fixtures.serverclass.thread.get_ephemeral_host', return_value=sentinel.host)
def test_init_with_random_hostname(mock_get_ephemeral_host):
    ts = ThreadServer(sentinel.cmd,
                      sentinel.get_args,
                      sentinel.env,
                      sentinel.workspace,
                      cwd=sentinel.cwd,
                      random_hostname=True)

    mock_get_ephemeral_host.assert_called_once()
    assert ts.hostname == sentinel.host


@patch('pytest_server_fixtures.serverclass.thread.get_ephemeral_host')
def test_init_without_random_hostname(mock_get_ephemeral_host):
    ts = ThreadServer(sentinel.cmd,
                      sentinel.get_args,
                      sentinel.env,
                      sentinel.workspace,
                      cwd=sentinel.cwd,
                      random_hostname=False)

    mock_get_ephemeral_host.assert_not_called()
    assert ts.hostname == CONFIG.fixture_hostname





'''
Created on 25 Apr 2012

@author: eeaston

'''
from __future__ import absolute_import
import socket

import pytest

from pytest_server_fixtures import CONFIG
from pytest_fixture_config import requires_config

from .base2 import TestServerV2


def _redis_server(request):
    """ Does the redis server work, this is used within different scoped
        fixtures.
    """
    test_server = RedisTestServer()
    request.addfinalizer(lambda p=test_server: p.teardown())
    test_server.start()
    return test_server


@pytest.fixture(scope='function')
@requires_config(CONFIG, ['redis_executable'])
def redis_server(request):
    """ Function-scoped Redis server in a local thread.

        Attributes
        ----------
        api: (``redis.Redis``)   Redis client API connected to this server
        .. also inherits all attributes from the `workspace` fixture
    """
    return _redis_server(request)


@pytest.fixture(scope='session')
@requires_config(CONFIG, ['redis_executable'])
def redis_server_sess(request):
    """ Same as redis_server fixture, scoped for test session
    """
    return _redis_server(request)


class RedisTestServer(TestServerV2):
    """This will look for 'redis_executable' in configuration and use as the
    redis-server to run.
    """

    def __init__(self, db=0, delete=True, **kwargs):
        global redis
        import redis

        super(RedisTestServer, self).__init__(delete=delete, **kwargs)
        self.db = db
        self._api = None
        self._port = self._get_port(6379)

    @property
    def api(self):
        if not self.hostname:
            raise "Redis not ready"
        if not self._api:
            self._api = redis.Redis(host=self.hostname, port=self.port, db=self.db)
        return self._api

    @property
    def cmd(self):
        return "redis-server"

    @property
    def cmd_local(self):
        return CONFIG.redis_executable

    def get_args(self, **kwargs):
        cmd = [
            "--port", str(self.port),
            "--timeout", "0",
            "--loglevel", "notice",
            "--databases", "1",
            "--maxmemory", "2gb",
            "--maxmemory-policy", "noeviction",
            "--appendonly", "no",
            "--slowlog-log-slower-than", "-1",
            "--slowlog-max-len", "1024",
        ]

        if 'hostname' in kwargs:
            cmd += ["--bind", "%s" % kwargs['hostname']]
        else:
            cmd += ["--bind", "0.0.0.0"]

        return cmd

    @property
    def image(self):
        return CONFIG.redis_image

    @property
    def port(self):
        return self._port

    def check_server_up(self):
        """ Ping the server
        """
        try:
            print("pinging Redis at %s:%s db %s" % (
                self.hostname, self.port, self.db
            ))
            return self.api.ping()
        except redis.ConnectionError as e:
            print("server not up yet (%s)" % e)
            return False



import os.path
from pytest import raises

from mock import patch


# patch out any changes you want to the Jenkins server here:
# These are once-for-all changes!

this_dir = os.path.dirname(__file__)
PLUGIN_REPO = os.path.join(this_dir, 'jenkins_plugins')


def test_load_plugins_fails_with_invalid_repo(jenkins_server_module):
    with raises(ValueError) as e:
        jenkins_server_module.load_plugins('junk_repo_name')
    assert str(e.value) == 'Plugin repository "junk_repo_name" does not exist'

def test_load_plugins_fails_with_invalid_plugin_name_as_string(jenkins_server_module):
    with raises(ValueError) as e:
        jenkins_server_module.load_plugins(PLUGIN_REPO, 'junk')
    assert str(e.value) == 'Plugin "junk" is not present in the repository'

def test_load_plugins_fails_with_invalid_plugin_name_as_list(jenkins_server_module):
    with raises(ValueError)as e:
        jenkins_server_module.load_plugins(PLUGIN_REPO, ['junk'])
    assert str(e.value) == 'Plugin "junk" is not present in the repository'

def test_load_plugins_fails_with_invalid_plugin_name_no_duplicates_in_error_msg(jenkins_server_module):
    with raises(ValueError) as e:
        jenkins_server_module.load_plugins(PLUGIN_REPO, ['junk', 'junk'])
    assert str(e.value) == 'Plugin "junk" is not present in the repository'

def test_load_plugins_fails_with_invalid_plugin_names_as_list(jenkins_server_module):
    with raises(ValueError)as e:
        jenkins_server_module.load_plugins(PLUGIN_REPO, ['zjunk', 'junk'])
    assert str(e.value) == "Plugins ['junk', 'zjunk'] are not present in the repository"

def test_load_plugins_loads_only_nominated_plugins(jenkins_server_module):
    with patch('pytest_server_fixtures.jenkins.shutil.copy') as mock_copy:
        jenkins_server_module.load_plugins(PLUGIN_REPO, 'notification')
        assert mock_copy.call_count == 1
        tup = mock_copy.call_args_list[0][0]
        assert tup[0].endswith('jenkins_plugins/notification.hpi')
        assert str(tup[1]) == os.path.join(jenkins_server_module.workspace, 'plugins/notification.hpi')

def test_load_plugins_loads_all_plugins(jenkins_server_module):
    with patch('pytest_server_fixtures.jenkins.shutil.copy') as mock_copy:
        jenkins_server_module.load_plugins(PLUGIN_REPO)
        assert mock_copy.call_count == len([x for x in os.listdir(PLUGIN_REPO)
                                            if x.endswith('.hpi')])

def test_jenkins_pre_server(jenkins_server_module):
    """ Creates template, creates the jenkins job
    """
    assert jenkins_server_module.check_server_up()



"""
Kubernetes server class implementation.
"""
from __future__ import absolute_import

import os
import logging
import uuid

from kubernetes import config
from kubernetes import client as k8sclient
from kubernetes.client.rest import ApiException
from retry import retry
from pytest_server_fixtures import CONFIG
from .common import (ServerClass,
                     merge_dicts,
                     ServerFixtureNotRunningException,
                     ServerFixtureNotTerminatedException)

log = logging.getLogger(__name__)

IN_CLUSTER = os.path.exists('/var/run/secrets/kubernetes.io/namespace')
fixture_namespace = CONFIG.k8s_namespace

if IN_CLUSTER:
    config.load_incluster_config()
    if not fixture_namespace:
        with open('/var/run/secrets/kubernetes.io/namespace', 'r') as f:
            fixture_namespace = f.read().strp()
        log.info("SERVER_FIXTURES_K8S_NAMESPACE is not set, using current namespace '%s'", fixture_namespace)

if CONFIG.k8s_local_test:
    log.info("====== Running K8S Server Class in Test Mode =====")
    config.load_kube_config()
    fixture_namespace = 'default'


class NotRunningInKubernetesException(Exception):
    """Thrown when code is not running as a Pod inside a Kubernetes cluster."""
    pass


class KubernetesServer(ServerClass):
    """Kubernetes server class."""

    def __init__(self,
                server_type,
                cmd,
                get_args,
                env,
                image,
                labels={}):
        super(KubernetesServer, self).__init__(cmd, get_args, env)

        if not fixture_namespace:
            raise NotRunningInKubernetesException()

        self._image = image
        self._labels = merge_dicts(labels, {
            'server-fixtures': 'kubernetes-server-fixtures',
            'server-fixtures/server-type': server_type,
            'server-fixtures/session-id': CONFIG.session_id,
        })

        self._v1api = k8sclient.CoreV1Api()

    def launch(self):
        try:
            log.debug('%s Launching pod' % self._log_prefix)
            self._create_pod()
            self._wait_until_running()
            log.debug('%s Pod is running' % self._log_prefix)
        except ApiException as e:
            log.warning('%s Error while launching pod: %s', self._log_prefix, e)
            raise

    def run(self):
        pass

    def teardown(self):
        self._delete_pod()
        # TODO: provide an flag to skip the wait to speed up the tests?
        self._wait_until_teardown()

    @property
    def is_running(self):
        try:
            return self._get_pod_status().phase == 'Running'
        except ApiException as e:
            if e.status == 404:
                # return false if pod does not exists
                return False
            raise

    @property
    def hostname(self):
        if not self.is_running:
            raise ServerFixtureNotRunningException()
        return self._get_pod_status().pod_ip

    @property
    def namespace(self):
        return fixture_namespace

    @property
    def labels(self):
        return self._labels

    def _get_pod_spec(self):
        container = k8sclient.V1Container(
            name='fixture',
            image=self._image,
            command=self._get_cmd(),
            env=[k8sclient.V1EnvVar(name=k, value=v) for k, v in self._env.iteritems()],
        )

        return k8sclient.V1PodSpec(
            containers=[container]
        )

    def _create_pod(self):
        try:
            pod = k8sclient.V1Pod()
            pod.metadata = k8sclient.V1ObjectMeta(name=self.name, labels=self._labels)
            pod.spec = self._get_pod_spec()
            self._v1api.create_namespaced_pod(namespace=self.namespace, body=pod)
        except ApiException as e:
            log.error("%s Failed to create pod: %s", self._log_prefix, e.reason)
            raise

    def _delete_pod(self):
        try:
            body = k8sclient.V1DeleteOptions()
            # delete the pod without waiting
            body.grace_period_seconds = 1
            self._v1api.delete_namespaced_pod(namespace=self.namespace, name=self.name, body=body)
        except ApiException as e:
            log.error("%s Failed to delete pod: %s", self._log_prefix, e.reason)

    def _get_pod_status(self):
        try:
            resp = self._v1api.read_namespaced_pod_status(namespace=self.namespace, name=self.name)
            return resp.status
        except ApiException as e:
            log.error("%s Failed to read pod status: %s", self._log_prefix, e.reason)
            raise

    @retry(ServerFixtureNotRunningException, tries=28, delay=1, backoff=2, max_delay=10)
    def _wait_until_running(self):
        log.debug("%s Waiting for pod status to become running", self._log_prefix)
        if not self.is_running:
            raise ServerFixtureNotRunningException()

    @retry(ServerFixtureNotTerminatedException, tries=28, delay=1, backoff=2, max_delay=10)
    def _wait_until_teardown(self):
        try:
            self._get_pod_status()
            # waiting for pod to be deleted (expect ApiException with status 404)
            raise ServerFixtureNotTerminatedException()
        except ApiException as e:
            if e.status == 404:
                return
            raise

    @property
    def _log_prefix(self):
        return "[K8S %s:%s]" % (self.namespace, self.name)



"""
Common utils for all serverclasses
"""
import os
import threading

from pytest_server_fixtures import CONFIG
from pytest_server_fixtures.util import get_random_id

SERVER_ID_LEN = 8

def merge_dicts(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z


def is_debug():
    return 'DEBUG' in os.environ and os.environ['DEBUG'] == '1'


class ServerFixtureNotRunningException(Exception):
    """Thrown when a kubernetes pod is not in running state."""
    pass


class ServerFixtureNotTerminatedException(Exception):
    """Thrown when a kubernetes pod is still running."""
    pass


class ServerClass(threading.Thread):
    """Example interface for ServerClass."""

    def __init__(self,
                 cmd,
                 get_args,
                 env):
        """
        Initialise the server class.
        Server fixture will be started here.
        """
        super(ServerClass, self).__init__()

        # set serverclass thread to a daemon thread
        self.daemon = True

        self._id = get_random_id(SERVER_ID_LEN)
        self._cmd = cmd
        self._get_args = get_args
        self._env = env or {}

    def run(self):
        """In a new thread, wait for the server to return."""
        raise NotImplementedError("Concrete class should implement this")

    def launch(self):
        """Start the server."""
        raise NotImplementedError("Concrete class should implement this")

    def teardown(self):
        """Kill the server."""
        raise NotImplementedError("Concrete class should implement this")

    @property
    def is_running(self):
        """Tell if the server is running."""
        raise NotImplementedError("Concrete class should implement this")

    @property
    def hostname(self):
        """Get server's hostname."""
        raise NotImplementedError("Concrete class should implement this")

    @property
    def name(self):
        return "server-fixtures-%s-%s" % (CONFIG.session_id, self._id)




import socket
import os

from pytest_fixture_config import Config
from .util import get_random_id


SESSION_ID_LEN = 8

class FixtureConfig(Config):
    __slots__ = (
        'java_executable',
        'jenkins_url',
        'jenkins_war',
        'jenkins_image',
        'minio_executable',
        'minio_image',
        'mongo_bin',
        'mongo_image',
        'redis_executable',
        'redis_image',
        'rethink_executable',
        'rethink_image',
        'httpd_executable',
        'httpd_image',
        'httpd_modules',
        'fixture_hostname',
        'xvfb_executable',
        'disable_proxy',
        'server_class',
        'session_id',
        'k8s_namespace',
        'k8s_local_test'
    )

# Default values for system resource locations - patch this to change defaults
DEFAULT_SERVER_FIXTURES_HOSTNAME = socket.gethostbyname(socket.gethostname())
DEFAULT_SERVER_FIXTURES_SESSION_ID = get_random_id(SESSION_ID_LEN)
DEFAULT_SERVER_FIXTURES_DISABLE_HTTP_PROXY = True
DEFAULT_SERVER_FIXTURES_SERVER_CLASS = 'thread'
DEFAULT_SERVER_FIXTURES_K8S_NAMESPACE = None
DEFAULT_SERVER_FIXTURES_K8S_LOCAL_TEST = False
DEFAULT_SERVER_FIXTURES_JAVA = 'java'
DEFAULT_SERVER_FIXTURES_JENKINS_URL = 'http://acmejenkins.example.com'
DEFAULT_SERVER_FIXTURES_JENKINS_WAR = '/usr/share/jenkins/jenkins.war'
DEFAULT_SERVER_FIXTURES_JENKINS_IMAGE = 'jenkins/jenkins:2.138.3-alpine'
DEFAULT_SERVER_FIXTURES_MINIO = 'minio'
DEFAULT_SERVER_FIXTURES_MINIO_IMAGE = 'minio/minio:latest'
DEFAULT_SERVER_FIXTURES_MONGO_BIN = 'mongod'
DEFAULT_SERVER_FIXTURES_MONGO_IMAGE = 'mongo:3.6'
DEFAULT_SERVER_FIXTURES_REDIS = 'redis-server'
DEFAULT_SERVER_FIXTURES_REDIS_IMAGE = 'redis:5.0.2-alpine'
DEFAULT_SERVER_FIXTURES_RETHINK = 'rethinkdb'
DEFAULT_SERVER_FIXTURES_RETHINK_IMAGE = 'rethinkdb:2.3.6'
DEFAULT_SERVER_FIXTURES_HTTPD = 'apache2'
DEFAULT_SERVER_FIXTURES_HTTPD_IMAGE = 'httpd:2.4.37'
DEFAULT_SERVER_FIXTURES_HTTPD_MODULES = '/usr/lib/apache2/modules'
DEFAULT_SERVER_FIXTURES_XVFB = 'Xvfb'


# Global config for finding system resources.
CONFIG = FixtureConfig(
    # Not using localhost here in case we are being used in a cluster-type job
    fixture_hostname=os.getenv('SERVER_FIXTURES_HOSTNAME', DEFAULT_SERVER_FIXTURES_HOSTNAME),
    disable_proxy=os.getenv('SERVER_FIXTURES_DISABLE_HTTP_PROXY', DEFAULT_SERVER_FIXTURES_DISABLE_HTTP_PROXY),
    server_class=os.getenv('SERVER_FIXTURES_SERVER_CLASS', DEFAULT_SERVER_FIXTURES_SERVER_CLASS),
    k8s_namespace=os.getenv('SERVER_FIXTURES_K8S_NAMESPACE', DEFAULT_SERVER_FIXTURES_K8S_NAMESPACE),
    k8s_local_test=os.getenv('SERVER_FIXTURES_K8S_LOCAL_TEST', DEFAULT_SERVER_FIXTURES_K8S_LOCAL_TEST),
    session_id=os.getenv('SERVER_FIXTURES_SESSION_ID', DEFAULT_SERVER_FIXTURES_SESSION_ID),
    java_executable=os.getenv('SERVER_FIXTURES_JAVA', DEFAULT_SERVER_FIXTURES_JAVA),
    jenkins_war=os.getenv('SERVER_FIXTURES_JENKINS_WAR', DEFAULT_SERVER_FIXTURES_JENKINS_WAR),
    jenkins_image=os.getenv('SERVER_FIXTURES_JENKINS_IMAGE', DEFAULT_SERVER_FIXTURES_JENKINS_IMAGE),
    minio_executable=os.getenv('SERVER_FIXTURES_MINIO', DEFAULT_SERVER_FIXTURES_MINIO),
    minio_image=os.getenv('SERVER_FIXTURES_MINIO_IMAGE', DEFAULT_SERVER_FIXTURES_MINIO_IMAGE),
    mongo_bin=os.getenv('SERVER_FIXTURES_MONGO_BIN', DEFAULT_SERVER_FIXTURES_MONGO_BIN),
    mongo_image=os.getenv('SERVER_FIXTURES_MONGO_IMAGE', DEFAULT_SERVER_FIXTURES_MONGO_IMAGE),
    redis_executable=os.getenv('SERVER_FIXTURES_REDIS', DEFAULT_SERVER_FIXTURES_REDIS),
    redis_image=os.getenv('SERVER_FIXTURES_REDIS_IMAGE', DEFAULT_SERVER_FIXTURES_REDIS_IMAGE),
    rethink_executable=os.getenv('SERVER_FIXTURES_RETHINK', DEFAULT_SERVER_FIXTURES_RETHINK),
    rethink_image=os.getenv('SERVER_FIXTURES_RETHINK_IMAGE', DEFAULT_SERVER_FIXTURES_RETHINK_IMAGE),
    httpd_executable=os.getenv('SERVER_FIXTURES_HTTPD', DEFAULT_SERVER_FIXTURES_HTTPD),
    httpd_modules=os.getenv('SERVER_FIXTURES_HTTPD_MODULES', DEFAULT_SERVER_FIXTURES_HTTPD_MODULES),
    httpd_image=os.getenv('SERVER_FIXTURES_HTTPD_IMAGE', DEFAULT_SERVER_FIXTURES_HTTPD_IMAGE),
    xvfb_executable=os.getenv('SERVER_FIXTURES_XVFB', DEFAULT_SERVER_FIXTURES_XVFB),
)



import socket
import uuid
import logging

import pytest

from pytest_server_fixtures import CONFIG
from pytest_fixture_config import requires_config

from .base2 import TestServerV2


log = logging.getLogger(__name__)
rethinkdb = None


def _rethink_server(request):
    """ This does the actual work - there are several versions of this used
        with different scopes.
    """
    test_server = RethinkDBServer()
    request.addfinalizer(lambda p=test_server: p.teardown())
    test_server.start()
    return test_server


@pytest.fixture(scope='function')
@requires_config(CONFIG, ['rethink_executable'])
def rethink_server(request):
    """ Function-scoped RethinkDB server in a local thread.

        Attributes
        ----------
        conn: (``rethinkdb.Connection``)  Connection to this server instance
        .. also inherits all attributes from the `workspace` fixture

    """
    return _rethink_server(request)


@pytest.fixture(scope='session')
@requires_config(CONFIG, ['rethink_executable'])
def rethink_server_sess(request):
    """ Same as rethink_server fixture, scoped as session instead.
    """
    return _rethink_server(request)


@pytest.yield_fixture(scope="function")
def rethink_unique_db(rethink_server_sess):
    """ Starts up a session-scoped server, and returns a connection to
        a unique database for the life of a single test, and drops it after
    """
    dbid = uuid.uuid4().hex
    conn = rethink_server_sess.conn
    rethinkdb.db_create(dbid).run(conn)
    conn.use(dbid)
    yield conn
    rethinkdb.db_drop(dbid).run(conn)


@pytest.yield_fixture(scope="module")
def rethink_module_db(rethink_server_sess):
    """ Starts up a module-scoped server, and returns a connection to
        a unique database for all the tests in one module.
        Drops the database after module tests are complete.
    """
    dbid = uuid.uuid4().hex
    conn = rethink_server_sess.conn
    log.info("Making database")
    rethinkdb.db_create(dbid).run(conn)
    conn.use(dbid)
    yield conn
    log.info("Dropping database")
    rethinkdb.db_drop(dbid).run(conn)


@pytest.fixture(scope="module")
def rethink_make_tables(request, rethink_module_db):
    """ Module-scoped fixture that creates all tables specified in the test
        module attribute FIXTURE_TABLES.

    """
    reqd_table_list = getattr(request.module, 'FIXTURE_TABLES')
    log.debug("Do stuff before all module tests with {0}".format(reqd_table_list))
    conn = rethink_module_db
    for table_name, primary_key in reqd_table_list:
        try:
            rethinkdb.db(conn.db).table_create(table_name,
                                               primary_key=primary_key,
                                               ).run(conn)
            log.info('Made table "{0}" with key "{1}"'
                     .format(table_name, primary_key))
        except rethinkdb.errors.RqlRuntimeError as err:
            log.debug('Table "{0}" not made: {1}'.format(table_name, err.message))


@pytest.yield_fixture(scope="function")
def rethink_empty_db(request, rethink_module_db, rethink_make_tables):
    """ Function-scoped fixture that will empty all the tables defined
        for the `rethink_make_tables` fixture.

        This is a useful approach, because of the long time taken to
        create a new RethinkDB table, compared to the time to empty one.
    """
    tables_to_emptied = (table[0] for table
                         in getattr(request.module, 'FIXTURE_TABLES'))
    conn = rethink_module_db

    for table_name in tables_to_emptied:
        rethinkdb.db(conn.db).table(table_name).delete().run(conn)
        log.debug('Emptied "{0}" before test'.format(table_name))
    yield conn


class RethinkDBServer(TestServerV2):
    random_hostname = False

    def __init__(self, **kwargs):
        # defer loading of rethinkdb
        global rethinkdb
        import rethinkdb

        super(RethinkDBServer, self).__init__(**kwargs)
        self._driver_port = self._get_port(28015)
        self._cluster_port = self._get_port(29015)
        self._http_port = self._get_port(8080)
        self.db = None


    @property
    def cmd(self):
        return "rethindb"

    @property
    def cmd_local(self):
        return CONFIG.rethink_executable

    def get_args(self, **kwargs):
        cmd = [
            '--driver-port', str(self.port),
            '--http-port', str(self.http_port),
            '--cluster-port', str(self.cluster_port),
        ]

        if 'hostname' in kwargs:
            cmd += ['--bind', kwargs['hostname']]
        else:
            cmd += ['--bind', 'all']

        if 'workspace' in kwargs:
            cmd += ['--directory', str(kwargs['workspace'] / 'db')]

        return cmd

    @property
    def image(self):
        return CONFIG.rethink_image

    @property
    def port(self):
        return self._driver_port

    @property
    def cluster_port(self):
        return self._cluster_port

    @property
    def http_port(self):
        return self._http_port

    def check_server_up(self):
        """Test connection to the server."""
        log.info("Connecting to RethinkDB at {0}:{1}".format(
            self.hostname, self.port))
        try:
            self.conn = rethinkdb.connect(host=self.hostname,
                                          port=self.port, db='test')
            return True
        except rethinkdb.RqlDriverError as err:
            log.warn(err)
        return False



import psutil

def test_start_and_stop(httpd_server):
    assert httpd_server.check_server_up()
    pid = int((httpd_server.workspace / 'run' / 'httpd.pid').text())
    httpd_server.kill()
    assert not httpd_server.check_server_up()
    still_running = [i for i in psutil.process_iter()
                     if i.pid == pid or i.ppid == pid]
    assert not still_running


def test_logs(httpd_server):
    files = [i.basename() for i in httpd_server.log_dir.files()]
    for log in ('access.log', 'error.log'):
        assert log in files


def test_get_from_document_root(httpd_server):
    hello = httpd_server.document_root / 'hello.txt'
    hello.write_text('Hello World!')
    response = httpd_server.get('hello.txt')
    assert response.status_code == 200
    assert response.text == 'Hello World!'



import os
import tempfile
import shutil
import subprocess
import time
import errno
import logging
import getpass

import pytest

from pytest_server_fixtures import CONFIG
from pytest_fixture_config import yield_requires_config

from .base2 import TestServerV2

log = logging.getLogger(__name__)


def _mongo_server():
    """ This does the actual work - there are several versions of this used
        with different scopes.
    """
    test_server = MongoTestServer()
    try:
        test_server.start()
        yield test_server
    finally:
        test_server.teardown()


@pytest.yield_fixture(scope='function')
@yield_requires_config(CONFIG, ['mongo_bin'])
def mongo_server():
    """ Function-scoped MongoDB server started in a local thread.
        This also provides a temp workspace.
        We tear down, and cleanup mongos at the end of the test.

        For completeness, we tidy up any outstanding mongo temp directories
        at the start and end of each test session

        Attributes
        ----------
        api (`pymongo.MongoClient`)  : PyMongo Client API connected to this server
        .. also inherits all attributes from the `workspace` fixture
    """
    for server in _mongo_server():
        yield server


@pytest.yield_fixture(scope='session')
@yield_requires_config(CONFIG, ['mongo_bin'])
def mongo_server_sess():
    """ Same as mongo_server fixture, scoped as session instead.
    """
    for server in _mongo_server():
        yield server


@pytest.yield_fixture(scope='class')
@yield_requires_config(CONFIG, ['mongo_bin'])
def mongo_server_cls(request):
    """ Same as mongo_server fixture, scoped for test classes.
    """
    for server in _mongo_server():
        request.cls.mongo_server = server
        yield server


@pytest.yield_fixture(scope='module')
@yield_requires_config(CONFIG, ['mongo_bin'])
def mongo_server_module():
    """ Same as mongo_server fixture, scoped for test modules.
    """
    for server in _mongo_server():
        yield server


class MongoTestServer(TestServerV2):

    def __init__(self, delete=True, **kwargs):
        super(MongoTestServer, self).__init__(delete=delete, **kwargs)
        self._port = self._get_port(27017)
        self.api = None

    @property
    def cmd(self):
        return 'mongod'

    @property
    def cmd_local(self):
        return CONFIG.mongo_bin

    def get_args(self, **kwargs):
        cmd = [
            '--port=%s' % self.port,
            '--nounixsocket',
            '--syncdelay=0',
            '--nojournal',
            '--quiet',
        ]

        if 'hostname' in kwargs:
            cmd.append('--bind_ip=%s' % kwargs['hostname'])
        else:
            cmd.append('--bind_ip=0.0.0.0')
        if 'workspace' in kwargs:
            cmd.append('--dbpath=%s' % str(kwargs['workspace']))

        return cmd

    @property
    def image(self):
        return CONFIG.mongo_image

    @property
    def port(self):
        return self._port

    def check_server_up(self):
        """Test connection to the server."""
        import pymongo
        from pymongo.errors import AutoReconnect, ConnectionFailure

        log.info("Connecting to Mongo at %s:%s" % (self.hostname, self.port))
        try:
            self.api = pymongo.MongoClient(self.hostname, self.port,
                                           serverselectiontimeoutms=200)
            self.api.list_database_names()
            # Configure the client with default timeouts in case the server goes slow
            self.api = pymongo.MongoClient(self.hostname, self.port)
            return True
        except (AutoReconnect, ConnectionFailure) as e:
            pass
        return False

    def teardown(self):
        if self.api:
            self.api.close()
            self.api = None
        super(MongoTestServer, self).teardown()



from mock import sentinel, patch, Mock

from pytest_server_fixtures.serverclass.docker import DockerServer

@patch('pytest_server_fixtures.serverclass.docker.ServerClass.__init__')
def test_init(mock_init):
    s = DockerServer(sentinel.server_type,
                     sentinel.cmd,
                     sentinel.get_args,
                     sentinel.env,
                     sentinel.image)

    mock_init.assert_called_with(sentinel.cmd,
                                 sentinel.get_args,
                                 sentinel.env)




import pytest

def test_postgres_server(postgres_server_sess):
    conn = postgres_server_sess.connect('integration')
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE test (id serial PRIMARY KEY, num integer, data varchar);")
    cursor.execute("INSERT INTO test (num, data) VALUES (%s, %s)", (100, "abc'def"))
    cursor.execute("SELECT * FROM test;")
    assert cursor.fetchone() == (1, 100, "abc'def")





""" Base classes for all server fixtures.
"""
import hashlib
import os
import signal
import socket
import subprocess
import sys
import threading
import time
import traceback
from datetime import datetime
import logging
import random
import errno

from six import string_types

from pytest_server_fixtures import CONFIG
from pytest_shutil.workspace import Workspace

log = logging.getLogger(__name__)

OSX = sys.platform == 'darwin'

_SESSION_HOST = None

class ServerNotDead(Exception):
    pass


def get_ephemeral_host(cached=True, regen_cache=False):
    """
    Returns a random IP in the 127.0.0.0/24. This decreases the likelihood of races for ports by 255^3.

    Parameters
    ----------
    cached : ``bool``
        if True, use a cached return value. This is to preserve behaviour as many clients rely on the hostname
        to remain the same during a test session.

    regen_cache: ``bool``
        if True, regenerate the cached value.
    """
    global _SESSION_HOST

    if cached and not regen_cache and _SESSION_HOST:
        return _SESSION_HOST

    # MacOS / OSX does not support loopback ip addresses other than
    #  127.0.0.1 unless they are manually configured (unlike linux)
    if OSX:
        res = '127.0.0.1'
    else:
        res = '127.{}.{}.{}'.format(random.randrange(1, 255),
                                    random.randrange(1, 255),
                                    random.randrange(2, 255),)
    if regen_cache or not _SESSION_HOST:
        _SESSION_HOST = res
    return res


def get_ephemeral_port(port=0, host=None, cache_host=True):
    """
    Get an ephemeral socket at random from the kernel.

    Parameters
    ----------
    port: `str`
        If specified, use this port as a base and the next free port after that base will be returned.
    host: `str`
        If specified, use this host. Otherwise it will use a temporary IP in the 127.0.0.0/24 range
    cache_host: `bool`
        If True, the generated host is cached. This has no effect if you specify the host.

    Returns
    -------
    Available port to use
    """
    if host is None:
        host = get_ephemeral_host(cached=cache_host)

    # Dynamic port-range:
    # * cat /proc/sys/net/ipv4/ip_local_port_range
    # 32768   61000
    if port == 0:
        port = random.randrange(1024, 32768)

    while True:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind((host, port))
            port = s.getsockname()[1]
            s.close()
            return port
        except socket.error:
            port = random.randrange(1024, 32768)


class ProcessReader(threading.Thread):
    def __init__(self, process, stream, stderr):
        self.stderr = stderr
        self.process = process
        self.stream = stream
        super(ProcessReader, self).__init__()
        self.setDaemon(True)

    def run(self):
        while self.process.poll() is None:
            l = self.stream.readline()
            if not isinstance(l, string_types):
                l = l.decode('utf-8')

            if l.strip():
                if self.stderr:
                    sys.stderr.writelines(l.strip() + "\n")
                else:
                    log.debug(l.strip())


class ServerThread(threading.Thread):
    """ Class for running the server in a thread """

    def __init__(self, hostname, port, run_cmd, run_stdin=None, env=None, cwd=None):
        threading.Thread.__init__(self)
        self.hostname = hostname
        self.port = port
        self.run_cmd = run_cmd
        self.run_stdin = run_stdin
        self.daemon = True
        self.exit = False
        self.env = env or dict(os.environ)
        self.cwd = cwd or os.getcwd()

        if 'DEBUG' in os.environ:
            self.p = subprocess.Popen(self.run_cmd, env=self.env, cwd=self.cwd,
                                      stdin=subprocess.PIPE if run_stdin else None)
        else:
            self.p = subprocess.Popen(self.run_cmd, env=self.env, cwd=self.cwd,
                                      stdin=subprocess.PIPE if run_stdin else None,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE)
            ProcessReader(self.p, self.p.stdout, False).start()
            ProcessReader(self.p, self.p.stderr, True).start()

    def run(self):
        log.debug("Running server: %s" % ' '.join(self.run_cmd))
        log.debug("CWD: %s" % self.cwd)
        try:
            if self.run_stdin:
                log.debug("STDIN: %s" % self.run_stdin)
                self.p.stdin.write(self.run_stdin.encode('utf-8'))
            if self.p.stdin:
                self.p.stdin.close()
            self.p.wait()
        except OSError:
            if not self.exit:
                traceback.print_exc()


class TestServer(Workspace):
    """
    Abstract class for creating a working dir and setting up a server instance in a thread.

    Parameters
    ----------
    preserve_sys_path: `bool`
        Preserve our full sys.path in PYTHONPATH to make sure the child process can still import dependencies
        installed during tests_require
    cache_host: `bool`
        Use cached ephemeral hostnames
    """
    server = None
    serverclass = ServerThread  # Child classes can set this to a different serverthread class

    random_port = True  # Use a random or fixed port number
    port_seed = 65535  # Used to seed port numbers if not random_port
    kill_signal = signal.SIGTERM

    # Number of seconds to wait between kill retries. Increase if the service takes a while to die
    kill_retry_delay = 1

    def __init__(self, workspace=None, delete=None, preserve_sys_path=False, cache_host=True, **kwargs):
        super(TestServer, self).__init__(workspace=workspace, delete=delete)
        self.hostname = kwargs.get('hostname') or get_ephemeral_host(cached=cache_host)
        self.port = kwargs.get('port') or self.get_port()
        # We don't know if the server is alive or dead at this point, assume alive
        self.dead = False
        self.env = kwargs.get('env')
        self.cwd = kwargs.get('cwd')

        if preserve_sys_path:
            # If the child class was installed as a test dependency, the python dist files might not
            # be properly installed and subprocesses won't find them. When this flag is set
            # we preserve our full sys.path in PYTHONPATH to make sure the child process can still
            # import things properly
            env = self.env or dict(os.environ)
            env['PYTHONPATH'] = os.pathsep.join(sys.path)
            self.env = env

    def start(self):
        self.kill()
        try:
            self.pre_setup()
            self.start_server(env=self.env)
            self.post_setup()
            self.save()
        except:
            self.teardown()
            raise

    def get_port(self):
        """
        Pick repeatable but semi-random port based on hashed username, and the server class.
        """
        if not self.random_port:
            return self.port_seed - int(hashlib.sha1((os.environ['USER']
                                                      + self.__class__.__name__).encode('utf-8')).hexdigest()[:3], 16)
        return get_ephemeral_port(host=self.hostname)

    def pre_setup(self):
        """ This should execute any setup required before starting the server
        """
        pass

    @property
    def run_cmd(self):
        """ Child classes should implement this to return the commands needed
            to start the server
        """
        raise NotImplementedError("Concrete class should implement this")

    @property
    def run_stdin(self):
        """ This is passed to the server as stdin
        """
        return None

    def post_setup(self):
        """ This should execute any setup required after starting the server
        """
        pass

    def check_server_up(self):
        """ This is called to see if the server is up
        """
        raise NotImplementedError("Concrete class should implement this")

    def wait_for_go(self, start_interval=0.1, retries_per_interval=3, retry_limit=28, base=2.0):
        """
        This is called to wait until the server has started running.

        Uses a binary exponential backoff algorithm to set wait interval
        between retries. This finds the happy medium between quick starting
        servers (e.g. in-memory DBs) while remaining useful for the slower
        starting servers (e.g. web servers).

        Parameters
        ----------
        start_interval: ``float``
            initial wait interval in seconds
        retries_per_interval: ``int``
            number of retries before increasing waiting time
        retry_limit: ``int``
            total number of retries to attempt before giving up
        base: ``float``
            backoff multiplier

        """
        if start_interval <= 0.0:
            raise ValueError('start interval must be positive!')

        interval = start_interval

        retry_count = retry_limit
        start_time = datetime.now()
        while retry_count > 0:
            for _ in range(retries_per_interval):
                log.debug('sleeping for %s before retrying (%d of %d)'
                      % (interval, ((retry_limit + 1) - retry_count), retry_limit))
                if self.check_server_up():
                    log.debug('waited %s for server to start successfully'
                          % str(datetime.now() - start_time))
                    return

                time.sleep(interval)
                retry_count -= 1
            interval *= base

        raise ValueError("Server failed to start up after waiting %s. Giving up!"
                         % str(datetime.now() - start_time))

    def start_server(self, env=None):
        """ Start the server instance.
        """
        log.debug("Starting Server on host %s port %s" % (self.hostname, self.port))
        self.server = self.serverclass(self.hostname, self.port, self.run_cmd, self.run_stdin,
                                       env=getattr(self, "env", env), cwd=self.cwd)
        self.server.start()
        self.wait_for_go()
        log.debug("Server now awake")
        self.dead = False

    def _find_and_kill(self, retries, signal):
        log.debug("Killing server running at {}:{} using signal {}".format(self.hostname, self.port, signal))
        for _ in range(retries):
            if OSX:
                netstat_cmd = "lsof -n -i:{} | grep LISTEN | awk '{{ print $2 }}'".format(self.port)
            else:
                netstat_cmd = ("netstat -anp 2>/dev/null | grep %s:%s | grep LISTEN | "
                            "awk '{ print $7 }' | cut -d'/' -f1" % (socket.gethostbyname(self.hostname), self.port))
            pids = [p.strip() for p in self.run(netstat_cmd, capture=True, cd='/').split('\n') if p.strip()]

            if not pids:
                # No PIDs remaining, server has died.
                break

            for pid in pids:
                try:
                    pid = int(pid)
                except ValueError:
                    log.error("Can't determine port, process shutting down or owned by someone else")
                else:
                    try:
                        os.kill(pid, signal)
                    except OSError as oe:
                        if oe.errno == errno.ESRCH:  # Process doesn't appear to exist.
                            log.error("For some reason couldn't find PID {} to kill.".format(p))
                        else:
                            raise
            time.sleep(self.kill_retry_delay)
        else:
            raise ServerNotDead("Server not dead after %d retries" % retries)

    def kill(self, retries=5):
        """Kill all running versions of this server.

        Just killing the thread.server pid isn't good enough, it may have spawned children.

        """
        # Prevent traceback printed when the server goes away as we kill it
        if self.server:
            self.server.exit = True

        if self.dead:
            return

        try:
            self._find_and_kill(retries, self.kill_signal)
        except ServerNotDead:
            log.error("Server not dead after %d retries, trying with SIGKILL" % retries)
        try:
            self._find_and_kill(retries, signal.SIGKILL)
        except ServerNotDead:
            log.error("Server still not dead, giving up")

    def teardown(self):
        """ Called when tearing down this instance, eg in a context manager
        """
        self.kill()
        super(TestServer, self).teardown()

    def save(self):
        """ Called to save any state that can be then restored using self.restore
        """
        pass

    def restore(self):
        """ Called to restore any state that was saved using using self.save
        """
        pass



import os
import subprocess
import time

from itertools import chain, repeat

from mock import patch
import pytest
from pytest import raises

from pytest_server_fixtures.xvfb import XvfbServer


def test_construct(xvfb_server):
        assert xvfb_server.display


def test_connect_client():
    with XvfbServer() as server:
        p = subprocess.Popen(['xdpyinfo', '-display', server.display],
                             env=dict(os.environ, XAUTHORITY=server.authfile), stdout=subprocess.PIPE)
        dpyinfo, _ = p.communicate()
        assert p.returncode == 0
        assert server.display in str(dpyinfo)


def test_terminates_on_last_client_exit():
    with XvfbServer() as server:
        subprocess.check_call(['xdpyinfo', '-display', server.display],
                              env=dict(os.environ, XAUTHORITY=server.authfile), stdout=open('/dev/null'))
        for _ in range(5):
            if server.process.poll() is not None:
                break
            time.sleep(0.1)  # wait up to 0.5 seconds for the server to terminate
        assert server.process.poll() == 0


def test_tries_to_find_free_server_num():
    with XvfbServer() as server1:
        with XvfbServer() as server2:
            assert server1.display != server2.display


def test_raises_if_fails_to_find_free_server_num():
    _exists = os.path.exists
    with patch('os.path.exists', new=lambda f: "-lock" in f or _exists(f)):
        with raises(RuntimeError) as ex:
            XvfbServer()
        assert 'Unable to find a free server number to start Xvfb' in str(ex)


def test_handles_unexpected_server_num_collision():
    with XvfbServer() as server1:
        from os.path import exists as real_exists
        with patch('os.path.exists') as mock_exists:
            side_effect_chain = chain([lambda _: False], repeat(real_exists))
            mock_exists.side_effect = lambda path: next(side_effect_chain)(path)
            with XvfbServer() as server2:
                assert server1.display != server2.display


def test_handles_unexpected_failure_to_start():
    with patch('pytest_server_fixtures.xvfb.XvfbServer.xvfb_command', 'false'):
        with raises(RuntimeError) as ex:
            XvfbServer()
        assert 'Failed to start Xvfb' in str(ex)



import os
import hashlib
import logging
import time
from datetime import datetime

from pytest_server_fixtures import CONFIG
from pytest_shutil.workspace import Workspace
from .base import get_ephemeral_port
from .serverclass import create_server

log = logging.getLogger(__name__)


class TeardownTestServerException(Exception):
    """Thrown when attempting to start an already teardown test server."""
    pass


class TestServerV2(Workspace):
    """Base class of a v2 test server."""
    random_port = True
    random_hostname = True
    port_seed = 65535

    def __init__(self, cwd=None, workspace=None, delete=None, server_class=CONFIG.server_class):
        """
        Initialise a test server.

        @param cwd: the current working directory
        @param workspace: where all files will be stored
        @param delete: whether to delete the workspace after teardown or not
        @param server_class: specify server class name (default from CONFIG.server_class)
        """
        super(TestServerV2, self).__init__(workspace=workspace, delete=delete)
        self._cwd = cwd or os.getcwd()
        self._server_class = server_class
        self._server = None
        self._teardown = False

    def start(self):
        """
        Start the test server.
        """
        if self._teardown:
            raise TeardownTestServerException()

        try:
            self._server = create_server(
                server_class=CONFIG.server_class,
                server_type=self.__class__.__name__,
                cmd=self.cmd,
                cmd_local=self.cmd_local,
                get_args=self.get_args,
                env=self.env,
                image=self.image,
                labels=self.labels,
                workspace=self.workspace,
                cwd=self._cwd,
                random_hostname=self.random_hostname,
            )

            if self._server_class == 'thread':
                self.pre_setup()

            self._server.launch()
            self._wait_for_go()
            log.debug("Server now awake")

            self.post_setup()
        except OSError as err:
            log.warning("Error when starting the test server.")
            log.debug(err)
            raise

    def teardown(self):
        """
        Stop the server and clean up all resources.
        """
        if self._teardown:
            log.debug("Server is already teardown, skipping")
            return

        if not self._server:
            log.debug("Server not started yet, skipping")
            return
        self._server.teardown()
        self._server = None

        super(TestServerV2, self).teardown()

        self._teardown = True

    def check_server_up(self):
        """
        Check if the server is up.
        """
        raise NotImplementedError("Concret class should implement this")

    @property
    def hostname(self):
        """
        Get the IP address of the server.
        """
        return self._server.hostname

    @property
    def port(self):
        """
        Get the port number of the server.
        """
        raise NotImplementedError("Concret class should implement this")

    @property
    def cwd(self):
        """
        Get the current working directory of the server.
        """
        return self._cwd

    @property
    def image(self):
        """
        Get the Docker image of the server.

        Only used when SERVER_FIXTURE_SERVER_CLASS is 'docker' or 'kubernetes'.
        """
        raise NotImplementedError("Concret class should implement this")

    @property
    def labels(self):
        """
        Extra labels to be added to the server fixture container.

        Only used when SERVER_FIXTURE_SERVER_CLASS is 'docker' or 'kubernetes'.
        """
        return dict()

    @property
    def env(self):
        """
        Get the environment variables for running the server fixture.
        """
        return dict()

    @property
    def cmd(self):
        """
        Get the command to run the server fixture.
        """
        raise NotImplementedError("Concrete class should implement this")

    @property
    def cmd_local(self):
        """
        Get the local command to run the server fixture.

        Only used when SERVER_FIXTURES_SERVER_CLASS is 'thread'.
        """
        return self.get_cmd()

    def get_args(self, hostname=None, workspace=None):
        """
        Get the arguments to run the server fixtures.

        @param hostname: hostname of the server
        @param workspace: workspace of the server
        """
        raise NotImplementedError("Concrete class should implement this")

    def pre_setup(self):
        """
        DEPRECATED

        Only used when SERVER_FIXTURE_SERVER_CLASS is 'thread'
        """
        pass

    def post_setup(self):
        """
        Set up step to be run after server is up.
        """
        pass

    def _wait_for_go(self, start_interval=0.1, retries_per_interval=3, retry_limit=28, base=2.0):
        """
        This is called to wait until the server has started running.

        Uses a binary exponential backoff algorithm to set wait interval
        between retries. This finds the happy medium between quick starting
        servers (e.g. in-memory DBs) while remaining useful for the slower
        starting servers (e.g. web servers).

        Parameters
        ----------
        start_interval: ``float``
            initial wait interval in seconds
        retries_per_interval: ``int``
            number of retries before increasing waiting time
        retry_limit: ``int``
            total number of retries to attempt before giving up
        base: ``float``
            backoff multiplier

        """
        if start_interval <= 0.0:
            raise ValueError('start interval must be positive!')

        interval = start_interval

        retry_count = retry_limit
        start_time = datetime.now()
        while retry_count > 0:
            for _ in range(retries_per_interval):
                log.debug('sleeping for %s before retrying (%d of %d)'
                      % (interval, ((retry_limit + 1) - retry_count), retry_limit))
                if self.check_server_up():
                    log.debug('waited %s for server to start successfully'
                          % str(datetime.now() - start_time))
                    return

                time.sleep(interval)
                retry_count -= 1
            interval *= base

        raise ValueError("Server failed to start up after waiting %s. Giving up!"
                         % str(datetime.now() - start_time))

    def _get_port(self, default_port):
        """
        Get a random or pseudo-random port based on config.
        """
        if self._server_class != 'thread':
            return default_port

        return (
            get_ephemeral_port() if self.random_port
            else self._get_pseudo_random_port())

    def _get_pseudo_random_port(self):
        """
        Get a pseudo random port based on port_seed,
        classname and current username.
        """
        sig = (os.environ['USER'] + self.__class__.__name__).encode('utf-8')
        return self.port_seed - int(hashlib.sha1(sig).hexdigest()[:3], 16)



"""
Thread server class implementation
"""
import logging
import os
import signal
import subprocess
import traceback
import time
import psutil

from retry import retry

from pytest_server_fixtures import CONFIG
from pytest_server_fixtures.base import get_ephemeral_host, ProcessReader, ServerNotDead, OSX
from .common import ServerClass, is_debug

log = logging.getLogger(__name__)


# ThreadServer will attempt to kill all child processes recursively.
KILL_RETRY_COUNT=15 # Total retry count to kill if not all child processes are terminated.
KILL_RETRY_WAIT_SECS=1 # Wait time between two retries
KILL_WAIT_SECS=5 # Time to wait for processes to terminate in a single retry.


class ProcessStillRunningException(Exception):
    pass


@retry(ProcessStillRunningException,
       tries=KILL_RETRY_COUNT,
       delay=KILL_RETRY_WAIT_SECS)
def _kill_all(procs, sig):
    log.debug("Killing %d processes with signal %s", len(procs), sig)
    for p in procs:
        p.send_signal(sig)

    log.debug("Waiting for %d processes to die", len(procs))
    gone, alive = psutil.wait_procs(procs, timeout=KILL_WAIT_SECS)

    if len(alive) == 0:
        log.debug("All processes are terminated")
        return

    log.warning("%d processes remainings: %s", len(alive), ",".join(alive))
    raise ProcessStillRunningException()


def _kill_proc_tree(pid, sig=signal.SIGKILL, timeout=None):
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    children.append(parent)
    log.debug("Killing process tree for %d (total_procs_to_kill=%d)", parent.pid, len(children))
    _kill_all(children, sig)


class ThreadServer(ServerClass):
    """Thread server class."""

    def __init__(self,
                 cmd,
                 get_args,
                 env,
                 workspace,
                 cwd=None,
                 random_hostname=True):
        super(ThreadServer, self).__init__(cmd, get_args, env)

        self.exit = False
        self._hostname = get_ephemeral_host() if random_hostname else CONFIG.fixture_hostname
        self._workspace = workspace
        self._cwd = cwd
        self._proc = None

    def launch(self):
        log.debug("Launching thread server.")

        run_cmd = [self._cmd] + self._get_args(
            hostname=self.hostname,
            workspace=self._workspace,
        )

        debug = is_debug()

        extra_args = dict()
        if debug:
            extra_args['stdout'] = subprocess.PIPE
            extra_args['stderr'] = subprocess.PIPE

        self._proc = subprocess.Popen(run_cmd, env=self._env, cwd=self._cwd, **extra_args)
        log.debug("Running server: %s", ' '.join(run_cmd))
        log.debug("CWD: %s", self._cwd)

        if debug:
            ProcessReader(self._proc, self._proc.stdout, False).start()
            ProcessReader(self._proc, self._proc.stderr, True).start()

        self.start()

    def run(self):
        """Run in thread"""
        try:
            self._proc.wait()
        except OSError:
            if not self.exit:
                traceback.print_exc()

    @property
    def is_running(self):
        """Check if the main process is still running."""
        # return False if the process is not started yet
        if not self._proc:
            return False
        # return False if there is a return code from the main process
        return self._proc.poll() is None

    @property
    def hostname(self):
        return self._hostname

    def teardown(self):
        if not self._proc:
            log.warning("No process is running, skip teardown.")
            return

        _kill_proc_tree(self._proc.pid)
        self._proc = None




from mock import create_autospec, sentinel, call, patch, Mock

from pytest_server_fixtures.base2 import TestServerV2 as _TestServerV2 # TODO: why as _TestServerV2?

def test_init():
    with patch('pytest_shutil.workspace.Workspace.__init__', autospec=True) as init:
        ts = _TestServerV2(cwd=sentinel.cwd,
                           workspace=sentinel.workspace,
                           delete=sentinel.delete,
                           server_class=sentinel.server_class)
        assert init.call_args_list == [call(ts, workspace=sentinel.workspace, delete=sentinel.delete)]
        assert ts._cwd == sentinel.cwd
        assert ts._server_class == sentinel.server_class




import os
import socket
import string
import logging

import pytest
try:
    from path import Path
except ImportError:
    from path import path as Path

from pytest_fixture_config import yield_requires_config
from pytest_server_fixtures import CONFIG

from .http import HTTPTestServer

log = logging.getLogger(__name__)


@pytest.yield_fixture(scope='function')
@yield_requires_config(CONFIG, ['httpd_executable', 'httpd_modules'])
def httpd_server():
    """ Function-scoped httpd server in a local thread.
    
        Methods
        -------
        get()   : Query url relative to the server root.
        ..        Parse as json and retry failures by default.
        post()  : Post payload to url relative to the server root.
        ..        Parse as json and retry failures by default.
    """
    test_server = HTTPDServer()
    test_server.start()
    yield test_server
    test_server.teardown()


class HTTPDServer(HTTPTestServer):
    port_seed = 65531

    cfg_modules_template = """
      LoadModule headers_module $modules/mod_headers.so
      LoadModule proxy_module $modules/mod_proxy.so
      LoadModule proxy_http_module $modules/mod_proxy_http.so
      LoadModule proxy_connect_module $modules/mod_proxy_connect.so
      LoadModule alias_module $modules/mod_alias.so
      LoadModule dir_module $modules/mod_dir.so
      LoadModule autoindex_module $modules/mod_autoindex.so
      <IfModule !mod_log_config.c>
          LoadModule log_config_module $modules/mod_log_config.so
      </IfModule>
      LoadModule mime_module $modules/mod_mime.so
      LoadModule authz_core_module $modules/mod_authz_core.so
    """

    cfg_mpm_template = """
      LoadModule mpm_prefork_module $modules/mod_mpm_prefork.so
      StartServers       1
      MinSpareServers    1
      MaxSpareServers   4
      ServerLimit      4
      MaxClients       4
      MaxRequestsPerChild  10000
    """

    cfg_template = """
      TypesConfig /etc/mime.types

      ServerRoot $server_root
      Listen $listen_addr
      PidFile $server_root/run/httpd.pid

      ErrorLog $log_dir/error.log
      LogFormat "%h %l %u %t \\"%r\\" %>s %b" common
      CustomLog $log_dir/access.log common
      LogLevel info

      $proxy_rules

      Alias / $document_root/

      <Directory $server_root>
          Options +Indexes
      </Directory>
    """

    def __init__(self, proxy_rules=None, extra_cfg='', document_root=None, log_dir=None, **kwargs):
        """ httpd Proxy Server

        Parameters
        ----------
        proxy_rules: `dict`
            { proxy_src: proxy_dest }. Eg   {'/downstream_url/' : server.uri}
        extra_cfg: `str`
            Any extra Apache config
        document_root : `str`
            Server document root, defaults to temporary workspace
        log_dir : `str`
            Server log directory, defaults to $(workspace)/logs
        """
        self.proxy_rules = proxy_rules if proxy_rules is not None else {}
        self.cfg_template = string.Template(self.cfg_modules_template +
                                            self.cfg_mpm_template +
                                            self.cfg_template +
                                            extra_cfg)

        # Always print debug output for this process
        os.environ['DEBUG'] = '1'

        # Discover externally accessable hostname so selenium can get to it
        kwargs['hostname'] = kwargs.get('hostname', socket.gethostbyname(os.uname()[1]))

        super(HTTPDServer, self).__init__(**kwargs)
        
        self.document_root = document_root or self.workspace
        self.document_root = Path(self.document_root)
        self.log_dir = log_dir or self.workspace / 'logs'
        self.log_dir = Path(self.log_dir)

    def pre_setup(self):
        """ Write out the config file
        """
        self.config = self.workspace / 'httpd.conf'
        rules = []
        for source in self.proxy_rules:
            rules.append("ProxyPass {0} {1}".format(source, self.proxy_rules[source]))
            rules.append("ProxyPassReverse {0} {1} \n".format(source, self.proxy_rules[source]))
        cfg = self.cfg_template.substitute(
            server_root=self.workspace,
            document_root=self.document_root,
            log_dir=self.log_dir,
            listen_addr="{host}:{port}".format(host=self.hostname, port=self.port),
            proxy_rules='\n'.join(rules),
            modules=CONFIG.httpd_modules,
        )
        self.config.write_text(cfg)
        log.debug("=========== HTTPD Server Config =============\n{}".format(cfg))

        # This is where it stores PID files
        (self.workspace / 'run').mkdir()
        if not os.path.exists(self.log_dir):
            self.log_dir.mkdir()

    @property
    def run_cmd(self):
        return [CONFIG.httpd_executable, '-f', self.config]



import pytest

from mock import sentinel, patch, Mock

from pytest_server_fixtures.serverclass.kubernetes import KubernetesServer

@pytest.mark.skip(reason="Need a way to run this test in Kubernetes")
@patch('pytest_server_fixtures.serverclass.docker.ServerClass.__init__')
def test_init(mock_init):
    s = KubernetesServer(sentinel.server_type,
                         sentinel.cmd,
                         sentinel.get_args,
                         sentinel.env,
                         sentinel.image)

    mock_init.assert_called_with(sentinel.cmd,
                                 sentinel.get_args,
                                 sentinel.env)

